import string

import window

class TextWindow(window.Window):
	def __init__(self):
		self.dirtydata = 0 ### or 1?
		# true if we need to re-layout text
		self.dirtyview = 0
		# true if we need to redraw window
		self.scrollchar = 0
		# scrollchar is the character position which should appear at the
		#   top of the window (absolute position in text)
		self.scrollline = 0
		# scrollline is the line number containing scrollchar (line number
		#   in layout array
		self.layout = []
		# array of tuples (startchar, endchar, text), each representing a
		#   screen line

	def getcommandcontexts(self):
		return [TextWindow] + window.Window.getcommandcontexts(self)

	def reinit(self):
		self.dirtydata = 1
		self.layout = []
		self.scrollline = 0
	
	def see_change(self, obj, sig, detail):
		if (sig == watcher.SIG_DATA):
			self.dirtydata = 1
			self.layout = []
			self.scrollline = 0
			if (self.scrollchar > len(self.data.body)):
				self.scrollchar = 0
				self.scrollline = 0

	def refresh(self):
		if (self.dirtydata):
			self.dirtydata = 0
			self.dirtyview = 1
	
			self.layout_text(self.scrollchar, self.height)
		
			ix = self.height-2
			lx = self.scrollline + ix
			lx = min(lx, len(self.layout)-1)
			if (lx >= 0):
				lin = self.layout[lx]
				if (len(self.data.body) > lin[1]):
					self.layout_text(self.layout[0][0])
					ix = max(0,
						len(self.layout)-(self.height-1))
					if (ix > self.scrollline):
						self.scrollline = ix
						self.scrollchar = self.layout[self.scrollline][0]
		
		if (self.dirtyview):
			self.dirtyview = 0
			self.body.erase()
			
			if (len(self.layout) == 0):
				# I don't know why this is necessary, but a totally unused
				# window seems to cause redraw problems
				###self.body.addstr(0, 0, chr(0240))
				return

			for ix in range(self.height-1):
				lx = self.scrollline + ix
				if (lx < 0):
					continue
				if (lx >= len(self.layout)):
					break
				lintr = self.layout[lx][2]
				self.body.addstr(ix, 0, lintr)
		
	def layout_text(self, startchar=None, minvislines=None):
		### do something with minvislines
		if (startchar==None):
			startchar = self.scrollchar
		st = self.data.body
		lenst = len(st)

		pos = st.rfind('\n', 0, startchar)
		if (pos < 0):
			startchar = 0
		else:
			startchar = pos+1
		
		lnum = 0
		cx = startchar

		while (1):
			if (cx >= lenst):
				break

			if (lnum >= len(self.layout)):
				assert (lnum == len(self.layout))
				(nextcx, lin) = layout_one_line(self, st, cx)
				self.layout.append(lin)
				lnum = lnum+1
			
			elif (cx < self.layout[lnum][0]):
				(nextcx, lin) = layout_one_line(self, st, cx)
				self.layout.insert(lnum, lin)
				if (self.scrollline >= lnum):
					self.scrollline = self.scrollline+1
				lnum = lnum+1
				
			elif (cx == self.layout[lnum][0]):
				lnum = lnum+1
				if (lnum < len(self.layout)):
					nextcx = self.layout[lnum][0]
				else:
					nextcx = lenst
					
			elif (lnum+1 < len(self.layout) and cx < self.layout[lnum+1][0]):
				self.layout.pop(lnum)
				if (self.scrollline > lnum):
					self.scrollline = self.scrollline-1
				nextcx = cx

			else:
				assert 0, (cx, lenst, lnum, len(self.layout), self.layout[lnum])
		
			cx = nextcx

def layout_one_line(win, st, cx):
	pos = st.find('\n', cx)
	nextcx = pos+1
	if (pos < 0):
		pos = len(st)
		nextcx = pos
	while (1):
		lin = st[cx:pos]
		lintr = string.join(
			map(text.displaychar.__getitem__, map(ord, lin)),
			'')
		if (len(lintr) < window.scrwidth
			or (pos-cx <= 1)):
			break
		newpos = min(pos, cx+window.scrwidth-1)
		if (newpos < pos):
			pos = newpos
			nextcx = pos
			newpos = st.rfind(' ', cx, pos)
			if (newpos >= 0):
				pos = newpos
				nextcx = pos+1
			continue
		newpos = st.rfind(' ', cx, pos)
		if (newpos >= 0):
			pos = newpos
			nextcx = pos+1
		else:
			pos = pos-1
			nextcx = pos
		continue

	return (nextcx, (cx, pos, lintr))

import watcher
import text
import execute

execute.registercontext(TextWindow, 'textwindow')

